package mac10.test;

import java.util.LinkedList;

import mac10.connection.MessageConstants;
import mac10.models.Cell;
import mac10.models.WorldModel;
import mac10.monitor.Monitor;
import mac10.monitor.NewMonitor;
import mac10.roles.AbstractRole;
import mac10.util.AStarNormal;
import mac10.util.AStarHeap;
import mac10.util.Node;

public class NavTest {
	WorldModel world;
	Monitor monitor;
	NewMonitor newmonitor;

	public NavTest() {
		world = new WorldModel(80, 80, 0, 0, 0, 0, 0, 0);
        newmonitor = new NewMonitor(world);
	}

	public void drawGreen() {
		for (int i = 0; i < world.getHeight(); i++) {
			for (int j = 0; j < world.getWidth(); j++) {
				Cell cell = new Cell(i, j, 2);

				world.forceUpdateModel(cell);
			}
		}
	}

	public void makeWall() {
		for (int i = 10; i < 40; i++) {
			Cell cell = new Cell(i, 25, 2);
			cell.setObstacle(true);
			world.forceUpdateModel(cell);
		}
	}

	public void makeFence() {
		for (int i = 30; i < 49; i++) {
			Cell cell = new Cell(i, 25, 2);
			cell.setFence(true);
			world.forceUpdateModel(cell);
		}
		Cell cell = new Cell(49, 25, 2);
		cell.setSwitch(true);
		world.forceUpdateModel(cell);
	}
	
	public void makeObstacles(int startx, int endx, int starty, int endy) {
		for (int x = startx; x <= endx; x++) {
			for (int y = starty; y <= endy; y++) {
				Cell cell = new Cell(x, y, 2);
				cell.setObstacle(true);
				world.forceUpdateModel(cell);
			}
		}
		
	}

	public void drawWaypoints(int x, int y, LinkedList<String> list) {
		Cell cell = new Cell(x, y, 1);
		cell.setEnemyCorral(true);
		world.forceUpdateModel(cell);
		for (String string : list) {
			// System.out.println("way "+x+" "+y);
			// System.out.flush();
			if (string.equals(MessageConstants.ACTION_NORTH)) {
				y--;
			}
			if (string.equals(MessageConstants.ACTION_NORTHEAST)) {
				x++;
				y--;
			}
			if (string.equals(MessageConstants.ACTION_EAST)) {
				x++;
			}
			if (string.equals(MessageConstants.ACTION_SOUTHEAST)) {
				x++;
				y++;
			}
			if (string.equals(MessageConstants.ACTION_WEST)) {
				x--;
			}
			if (string.equals(MessageConstants.ACTION_NORTHWEST)) {
				x--;
				y--;
			}
			if (string.equals(MessageConstants.ACTION_SOUTH)) {
				y++;
			}
			if (string.equals(MessageConstants.ACTION_SOUTHWEST)) {
				y++;
				x--;
			}
			cell = new Cell(x, y, 1);
			cell.setAllyCorral(true);
			world.forceUpdateModel(cell);
		}
	}

	public void randomTrees(double c) {

		for (int i = 0; i < world.getHeight(); i++) {
			for (int j = 0; j < world.getWidth(); j++) {

				double d = Math.random();
				if (d < c) {
					Cell cell = new Cell(i, j, 2);
					cell.setObstacle(true);
					world.forceUpdateModel(cell);
				}

			}
		}

	}

	public static void main(String[] args) throws Exception {
		NavTest test = new NavTest();

		test.drawGreen();
		//test.makeWall();
		// test.makeFence();
		test.makeObstacles(0, 75, 10, 10);
		test.makeObstacles(5, 79, 20, 20);
		test.makeObstacles(0, 75, 30, 30);
		test.makeObstacles(5, 79, 40, 40);
		test.makeObstacles(0, 75, 50, 50);
		test.makeObstacles(5, 79, 60, 60);
		test.makeObstacles(0, 75, 70, 70);
		for(int i = 10; i < 80; i+= 20) {
			for(int j = 0; j < 80; j+= 10) {
				test.makeObstacles(i, i, j, j+7);
			}
		}
		for(int i = 20; i < 80; i+= 20) {
			for(int j = 0; j < 80; j+= 10) {
				test.makeObstacles(i, i, j+3, j+9);
			}
		}

		LinkedList<String> list = new LinkedList<String>();

		// list = AbstractRole.nodeToActionStrings(astar.AStar());

		// test.randomTrees(0.1);
		// test.drawWaypoints(10, 10, list);
//		 test.randomTrees(0.1);

		// list = AbstractRole.nodeToActionStrings(n19);

		int startx = 0;
		int starty = 0;
		int goalx = 0;
		int goaly = 79;

		AStarNormal astar = new AStarNormal(test.world, new Cell(startx, starty, 0),
				new Cell(goalx, goaly, 0));
		AStarHeap astarheap = new AStarHeap(test.world, new Cell(startx,
				starty, 0), new Cell(goalx, goaly, 0));

		// astar.setAvoidFences(true);
		Node goal = astar.AStar();

		long time = System.currentTimeMillis();
		for (int i = 0; i < 1; i++)
			goal = astar.AStar();
		System.out.println("Astar: " + (System.currentTimeMillis() - time));

		time = System.currentTimeMillis();
		for (int i = 0; i < 1; i++)
			goal = astarheap.AStar(System.currentTimeMillis() + 1000000);
		System.out.println("AstarHeap: " + (System.currentTimeMillis() - time));

		list = AbstractRole.nodeToActionStrings(goal);
		// System.out.println(new
		// Explorer(test.world).findCellToOpenFence(test.world.getCell(10, 10),
		// test.world.getCell(35, 25)));
		test.drawWaypoints(startx, starty, list);

//		Monitor.initializeMonitor(test.world);
//		Monitor.drawWorldModel(test.world, 2);
		
		test.newmonitor.updateMonitor(test.world);
	}

}